/*
 *    for STM32F746 Discovery
 *
 *    Copyright (C) by Hasada & FCT Embedded System Research Circle
 */

/* Includes ------------------------------------------------------------------*/
#include "imgpro_main.h"
#include "device_defines.h"
#include "rk043fn48h.h"
#include <string.h>
#include <stdlib.h>

#include "lcd_param.h"

/* for startup photo */
#include "bg320x240_T1_bmp.c"
#include "bg320x240_T2_bmp.c"

void  syslog(unsigned long prio, const char *format, ...);
int  dly_tsk( unsigned long );
int  tslp_tsk( unsigned long );


void  userfunc_main( int btnnum );
void  userfunc_init( void );


/* SD-Card ------------------------------------------------------------------*/
#include "stm32f7xx_hal.h"
#include "stm32746g_discovery.h"
#include "stm32746g_discovery_sd.h"

/* FatFs includes component */
#include "ff_gen_drv.h"
#include "sd_diskio.h"

FATFS SDFatFs;  /* File system object for SD card logical drive */
FIL MyFile;     /* File object */
char SDPath[4]; /* SD card logical drive path */

unsigned char z_bmp_header[ DEF_BMP_HEADER_SIZE ] = {
  0x42, 0x4D, 0x36, 0x84, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00,
  0x00, 0x00, 0x40, 0x01, 0x00, 0x00, 0xF0, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x84, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};


void  imgpro_sdcard_init( void )
{
  BSP_SD_Init();
  if( FATFS_LinkDriver(&SD_Driver, SDPath) != 0 ) {
    syslog(3, "ERROR : FATFS_LinkDriver()");
    return;
  }
}


/*
 * check file status in sdcard
 *
 *  return :  0 : file exists
 *            1 : no file
 *           -1 : access error
 */
int  imgpro_sdcard_file_check( char *fname )
{
  FRESULT  ret;
  FILINFO  info;
  int  stat = -1;

  if( f_mount( &SDFatFs, (TCHAR const*)SDPath, 0 ) != FR_OK) {
    syslog(3, "ERROR : f_mount()");
    return  stat;
  }

  ret = f_stat( (const TCHAR*)fname, &info );
  if( ret == FR_OK ) {
    stat = 0;
  } else if( ret == FR_NO_FILE ) {
    stat = 1;
  }

  return  stat;
}


void  imgpro_sdcard_read( char *fname, unsigned char *buff, int size )
{
  FRESULT  res;  /* FatFs function common result code */
  uint32_t  bytesread = 0;   /* File read counts */

  if( f_mount( &SDFatFs, (TCHAR const*)SDPath, 0 ) != FR_OK) {
    syslog(3, "ERROR : f_mount()");
    return;
  }

  if( f_open( &MyFile, fname, FA_READ ) != FR_OK ) {
    syslog(3, "ERROR : f_open()");
    return;
  }

  res = f_read( &MyFile, (uint8_t *)buff, size, (UINT*)&bytesread);
  syslog( 3, "INFO : f_read() res=%d  size=%d", res, bytesread );

  f_close( &MyFile );
}


void  imgpro_sdcard_write( char *fname, unsigned char *buff, int size )
{
  FRESULT  ret;
  UINT  cnt;
  int  remain = size;
  int  outsz = 0x100;
  int  errcnt = 0;

  if( f_mount( &SDFatFs, (TCHAR const*)SDPath, 0 ) != FR_OK) {
    syslog(3, "ERROR : f_mount()");
    return;
  }

  if( f_open( &MyFile, fname, FA_CREATE_NEW | FA_WRITE ) != FR_OK ) {
    syslog(3, "ERROR : f_open()");
    return;
  }

  do {
    if( remain < outsz ) {
      outsz = remain;
    }

    ret = f_write( &MyFile, buff, outsz, &cnt );

    buff += cnt;
    remain -= cnt;
    if( cnt == 0 ) {
      errcnt++;
      if( errcnt < 100 ) {
        tslp_tsk( 2 );
      } else {
        syslog(3, "ERROR : f_write() ret=%d  remain=%d  cnt=%d", ret, remain, cnt);
        break;
      }
    } else if( cnt != outsz ) {
      tslp_tsk( 2 );
    }
  } while( remain > 0 );

  f_close( &MyFile );
  tslp_tsk( 10 );
}


/*
 * save photo data to BMP file in SD card
 * photo is 320x240 pixel of RGB565
 *
 *  photo : top address of photo data
 *  fname : save file name
 */
void  imgpro_write_bmp_file( unsigned char *photo, char *fname )
{
  unsigned char *bmp;
  unsigned char *src;
  unsigned char *s0, *s1;
  int  i;
  int  x, y;
  int  size;

  /* copy BMP header */
  bmp = (unsigned char *)DEF_LCD_USER_BMP1;
  src = z_bmp_header;
  for( i=0 ; i<DEF_BMP_HEADER_SIZE ; i++ ) {
      *bmp++ = *src++;
  }

  /* copy photo data, RGB565 to BMP:BGR888 */
  for( y=0 ; y<240 ; y++ ) {
    src = photo + 320 * (240-1-y) * 2;
    for( x=0 ; x<320 ; x++ ) {
      s0 = src++;
      s1 = src++;
      *bmp++ = (*s0 & 0x1f) << 3;                      /* B */
      *bmp++ = ((*s1 & 0x07)<<5) | ((*s0 & 0xe0)>>3);  /* G */
      *bmp++ = *s1 & 0xf1;                             /* R */
    }
  }

  size = DEF_BMP_HEADER_SIZE + 320 * 240 * 3;
  imgpro_sdcard_write( fname, (unsigned char *)DEF_LCD_USER_BMP1, size );
}
/* SD-Card ------------------------------------------------------------------*/


/*
 *  0: 320x240
 *  1: 480x272
 *  2: 480x320
 */
int  imgpro_gra_size = 0 ;

int  disp_current_lcd ;
int  push_button_kind ;

/** @addtogroup STM32F7xx_HAL_Examples
  * @{
  */

/** @addtogroup FMC_SDRAM
  * @{
  */

/* Private typedef -----------------------------------------------------------*/
/* Private define ------------------------------------------------------------*/
#define BUFFER_SIZE         ((uint32_t)0x1000)
#define WRITE_READ_ADDR     ((uint32_t)0x0800)

/* Private macro -------------------------------------------------------------*/
/* Private variables ---------------------------------------------------------*/
SDRAM_HandleTypeDef      hsdram;
FMC_SDRAM_TimingTypeDef  SDRAM_Timing;
FMC_SDRAM_CommandTypeDef command;

/* Read/Write Buffers */
uint32_t aTxBuffer[BUFFER_SIZE];
uint32_t aRxBuffer[BUFFER_SIZE];

/* Status variables */
__IO uint32_t uwWriteReadStatus = 0;

/* Counter index */
uint32_t uwIndex = 0;


uint32_t SystemCoreClock = 160000000;
const uint8_t AHBPrescTable[16] = {0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 6, 7, 8, 9};
const uint8_t APBPrescTable[8] = {0, 0, 0, 0, 1, 2, 3, 4};


/* Private function prototypes -----------------------------------------------*/
static void Error_Handler(void);
static void BSP_SDRAM_Initialization_Sequence(SDRAM_HandleTypeDef *hsdram, FMC_SDRAM_CommandTypeDef *Command);
static void Fill_Buffer(uint32_t *pBuffer, uint32_t uwBufferLenght, uint32_t uwOffset);
static void CPU_CACHE_Enable(void);


void  busyloop_msec( int );

void HAL_Delay(__IO uint32_t Delay)
{
  int  i ;
  for( i=0 ; i<Delay ; i++ ) {
    busyloop_msec( 1 );
  }
}


void  usrbtn_init_PB8( void )
{
    GPIO_InitTypeDef gpio_init_structure;

    __HAL_RCC_GPIOB_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_8;
    gpio_init_structure.Mode = GPIO_MODE_INPUT;
    gpio_init_structure.Pull = GPIO_PULLUP;
    gpio_init_structure.Speed = GPIO_SPEED_FAST;
    HAL_GPIO_Init( GPIOB, &gpio_init_structure );
}


uint32_t  usrbtn_getbtn_PB8( void )
{
    return HAL_GPIO_ReadPin( GPIOB, GPIO_PIN_8 );
}


void  usrbtn_init_2( void )
{
    GPIO_InitTypeDef gpio_init_structure;

    __HAL_RCC_GPIOB_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_4;
    gpio_init_structure.Mode = GPIO_MODE_INPUT;
    gpio_init_structure.Pull = GPIO_PULLUP;
    gpio_init_structure.Speed = GPIO_SPEED_FAST;
    HAL_GPIO_Init( GPIOB, &gpio_init_structure );
}


uint32_t  usrbtn_getbtn_2( void )
{
    return HAL_GPIO_ReadPin( GPIOB, GPIO_PIN_4 );
}


void  usrbtn_init_3( void )
{
    GPIO_InitTypeDef gpio_init_structure;

    __HAL_RCC_GPIOH_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_6;
    gpio_init_structure.Mode = GPIO_MODE_INPUT;
    gpio_init_structure.Pull = GPIO_PULLUP;
    gpio_init_structure.Speed = GPIO_SPEED_FAST;
    HAL_GPIO_Init( GPIOH, &gpio_init_structure );
}


uint32_t  usrbtn_getbtn_3( void )
{
    return HAL_GPIO_ReadPin( GPIOH, GPIO_PIN_6 );
}


/* Definition for USARTx clock resources */
#define USARTx                           USART6
#define USARTx_CLK_ENABLE()              __USART6_CLK_ENABLE()
#define USARTx_RX_GPIO_CLK_ENABLE()      __GPIOC_CLK_ENABLE()
#define USARTx_TX_GPIO_CLK_ENABLE()      __GPIOC_CLK_ENABLE()

/* Definition for USARTx Pins */
#define USARTx_TX_PIN                    GPIO_PIN_6
#define USARTx_TX_GPIO_PORT              GPIOC
#define USARTx_TX_AF                     GPIO_AF8_USART6
#define USARTx_RX_PIN                    GPIO_PIN_7
#define USARTx_RX_GPIO_PORT              GPIOC
#define USARTx_RX_AF                     GPIO_AF8_USART6


void  xxHAL_UART_MspInit( void )
{  
  GPIO_InitTypeDef  GPIO_InitStruct;
  
  /* Enable GPIO TX/RX clock */
  USARTx_TX_GPIO_CLK_ENABLE();
  USARTx_RX_GPIO_CLK_ENABLE();


  /* Enable USARTx clock */
  USARTx_CLK_ENABLE(); 
  
  /* UART TX GPIO pin configuration  */
  GPIO_InitStruct.Pin       = USARTx_TX_PIN;
  GPIO_InitStruct.Mode      = GPIO_MODE_AF_PP;
  GPIO_InitStruct.Pull      = GPIO_PULLUP;
  GPIO_InitStruct.Speed     = GPIO_SPEED_HIGH;
  GPIO_InitStruct.Alternate = USARTx_TX_AF;

  HAL_GPIO_Init(USARTx_TX_GPIO_PORT, &GPIO_InitStruct);

  /* UART RX GPIO pin configuration  */
  GPIO_InitStruct.Pin = USARTx_RX_PIN;
  GPIO_InitStruct.Alternate = USARTx_RX_AF;

  HAL_GPIO_Init(USARTx_RX_GPIO_PORT, &GPIO_InitStruct);
}


UART_HandleTypeDef UartHandle;


void  UART_Config( void )
{
  UartHandle.Instance        = USARTx;

  UartHandle.Init.BaudRate   = 115200;

  UartHandle.Init.WordLength = UART_WORDLENGTH_8B;
  UartHandle.Init.StopBits   = UART_STOPBITS_1;
  UartHandle.Init.Parity     = UART_PARITY_NONE;
  UartHandle.Init.HwFlowCtl  = UART_HWCONTROL_NONE;
  UartHandle.Init.Mode       = UART_MODE_TX_RX;
  UartHandle.AdvancedInit.AdvFeatureInit = UART_ADVFEATURE_NO_INIT;
  if(HAL_UART_Init(&UartHandle) != HAL_OK)
  {
    Error_Handler();
  }
}


/***********************************************************************/
#define SPIx                             SPI2
#define SPIx_CLK_ENABLE()                __HAL_RCC_SPI2_CLK_ENABLE()
#define SPIx_SCK_GPIO_CLK_ENABLE()       __HAL_RCC_GPIOI_CLK_ENABLE()
#define SPIx_MISO_GPIO_CLK_ENABLE()      __HAL_RCC_GPIOB_CLK_ENABLE()
#define SPIx_MOSI_GPIO_CLK_ENABLE()      __HAL_RCC_GPIOB_CLK_ENABLE()

#define SPIx_FORCE_RESET()               __HAL_RCC_SPI2_FORCE_RESET()
#define SPIx_RELEASE_RESET()             __HAL_RCC_SPI2_RELEASE_RESET()

/* Definition for SPIx Pins */
#define SPIx_SCK_PIN                     GPIO_PIN_1
#define SPIx_SCK_GPIO_PORT               GPIOI
#define SPIx_SCK_AF                      GPIO_AF5_SPI2
#define SPIx_MISO_PIN                    GPIO_PIN_14
#define SPIx_MISO_GPIO_PORT              GPIOB
#define SPIx_MISO_AF                     GPIO_AF5_SPI2
#define SPIx_MOSI_PIN                    GPIO_PIN_15
#define SPIx_MOSI_GPIO_PORT              GPIOB
#define SPIx_MOSI_AF                     GPIO_AF5_SPI2

SPI_HandleTypeDef SpiHandle;
DMA_HandleTypeDef hdma_tx;

void xxHAL_SPI_MspInit( void )
{
    GPIO_InitTypeDef  GPIO_InitStruct;

    /* Enable GPIO TX/RX clock */
    SPIx_SCK_GPIO_CLK_ENABLE();
    SPIx_MISO_GPIO_CLK_ENABLE();
    SPIx_MOSI_GPIO_CLK_ENABLE();
    /* Enable SPI clock */
    SPIx_CLK_ENABLE(); 
    
    /* SPI SCK GPIO pin configuration  */
    GPIO_InitStruct.Pin       = SPIx_SCK_PIN;
    GPIO_InitStruct.Mode      = GPIO_MODE_AF_PP;
    GPIO_InitStruct.Pull      = GPIO_PULLUP;
    GPIO_InitStruct.Speed     = GPIO_SPEED_HIGH;
    GPIO_InitStruct.Alternate = SPIx_SCK_AF;
    HAL_GPIO_Init(SPIx_SCK_GPIO_PORT, &GPIO_InitStruct);

    /* SPI MISO GPIO pin configuration  */
    GPIO_InitStruct.Pin = SPIx_MISO_PIN;
    GPIO_InitStruct.Alternate = SPIx_MISO_AF;
    HAL_GPIO_Init(SPIx_MISO_GPIO_PORT, &GPIO_InitStruct);

    /* SPI MOSI GPIO pin configuration  */
    GPIO_InitStruct.Pin = SPIx_MOSI_PIN;
    GPIO_InitStruct.Alternate = SPIx_MOSI_AF;
    HAL_GPIO_Init(SPIx_MOSI_GPIO_PORT, &GPIO_InitStruct);

  {
    GPIO_InitTypeDef gpio_init_structure;

    /** /CS = NSS **/
    __HAL_RCC_GPIOA_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_15;
    gpio_init_structure.Mode = GPIO_MODE_OUTPUT_PP;
    HAL_GPIO_Init( GPIOA, &gpio_init_structure );
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 1);   /* /CS:H */

    /** /RST **/
    __HAL_RCC_GPIOI_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_2;
    gpio_init_structure.Mode = GPIO_MODE_OUTPUT_PP;
    HAL_GPIO_Init( GPIOI, &gpio_init_structure );
    HAL_GPIO_WritePin(GPIOI, GPIO_PIN_2, 1); /* /RST:H */

    /** /CS2 **/
    /* set clock enable */
    RCC->AHB1ENR |= ( RCC_AHB1ENR_GPIOGEN
                    ) ;
    /** for CS2:PG7 **/
    GPIOG->MODER   &= ~(0x0000c000);
    GPIOG->MODER   |=  (0x00004000);
    GPIOG->OTYPER  &= ~(0x0080);
    GPIOG->OSPEEDR &= ~(0x0000c000);
    GPIOG->OSPEEDR |=  (0x00008000);
    GPIOG->PUPDR   &= ~(0x0000c000);
    GPIOG->BSRR = 0x0080;        /* /CS2:H */

    /** for check pin1 **/
    __HAL_RCC_GPIOI_CLK_ENABLE() ;
    gpio_init_structure.Pin = GPIO_PIN_3;
    gpio_init_structure.Mode = GPIO_MODE_OUTPUT_PP;
    HAL_GPIO_Init( GPIOI, &gpio_init_structure );
    HAL_GPIO_WritePin(GPIOI, GPIO_PIN_3, 1);   /* H */
    HAL_GPIO_WritePin(GPIOI, GPIO_PIN_3, 0);   /* L */
  }
}


void  SPI_Config( void )
{
  /* Set the SPI parameters */
  SpiHandle.Instance               = SPIx;
  //SpiHandle.Init.BaudRatePrescaler = SPI_BAUDRATEPRESCALER_256;
  SpiHandle.Init.BaudRatePrescaler = SPI_BAUDRATEPRESCALER_2;
  SpiHandle.Init.Direction         = SPI_DIRECTION_2LINES;
  SpiHandle.Init.CLKPhase          = SPI_PHASE_1EDGE;
  //SpiHandle.Init.CLKPhase          = SPI_PHASE_2EDGE;
  SpiHandle.Init.CLKPolarity       = SPI_POLARITY_HIGH;
  //SpiHandle.Init.CLKPolarity       = SPI_POLARITY_LOW;
  SpiHandle.Init.DataSize          = SPI_DATASIZE_8BIT;
  SpiHandle.Init.FirstBit          = SPI_FIRSTBIT_MSB;
  SpiHandle.Init.TIMode            = SPI_TIMODE_DISABLE;
  SpiHandle.Init.CRCCalculation    = SPI_CRCCALCULATION_DISABLE;
  SpiHandle.Init.CRCPolynomial     = 7;
  SpiHandle.Init.NSS               = SPI_NSS_SOFT;
  //SpiHandle.Init.NSS               = SPI_NSS_HARD_OUTPUT;
  SpiHandle.Init.Mode              = SPI_MODE_MASTER;

  HAL_Delay(5); 
  if(HAL_SPI_Init(&SpiHandle) != HAL_OK) {
    /* Initialization Error */
    Error_Handler();
  }

  HAL_GPIO_WritePin(GPIOI, GPIO_PIN_2, 1); /* /RST:H */
  HAL_Delay(100);
  HAL_GPIO_WritePin(GPIOI, GPIO_PIN_2, 0); /* /RST:L */
  HAL_Delay(100);
  HAL_GPIO_WritePin(GPIOI, GPIO_PIN_2, 1); /* /RST:H */
  HAL_Delay(100);
}


void  writecom16( uint16_t com )
{
  uint8_t  tmp[4];

 if( disp_current_lcd == 0 ) {
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 0);   /* /CS:L */
 } else {
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 0);   /* /CS2:L */
  GPIOG->BSRR = 0x0080 << 16 ; /* /CS2:L */
 }

  tmp[0] = (uint8_t)0x70;
  tmp[1] = (uint8_t)((com >> 8) & 0xff); 
  tmp[2] = (uint8_t)(com & 0xff); 
  HAL_SPI_Transmit(&SpiHandle, tmp, 3, 1000);

  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 1);   /* /CS:H */
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 1);   /* /CS2:H */
  GPIOG->BSRR = 0x0080;        /* /CS2:H */
}


void  writedat16( uint16_t dat )
{
  uint8_t  tmp[4];

 if( disp_current_lcd == 0 ) {
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 0);   /* /CS:L */
 } else {
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 0);   /* /CS2:L */
  GPIOG->BSRR = 0x0080 << 16 ; /* /CS2:L */
 }

  tmp[0] = (uint8_t)0x72;
  tmp[1] = (uint8_t)((dat >> 8) & 0xff); 
  tmp[2] = (uint8_t)(dat & 0xff); 
  HAL_SPI_Transmit(&SpiHandle, tmp, 3, 1000);

  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 1);   /* /CS:H */
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 1);   /* /CS2:H */
  GPIOG->BSRR = 0x0080;        /* /CS2:H */
}


void  writereg16( uint16_t com, uint16_t dat )
{
  writecom16( com );
  writedat16( dat );
}


void  ili9328_init( void );
void  ili9328_setblock( uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1 );

void  ili9328_clr_framebuffer( void )
{
  unsigned long  *pnt = (unsigned long *)DEF_LCD_FB_TOP;
  unsigned long  *end = (unsigned long *)DEF_LCD_FB_BTM;

  while( pnt < end ) {
    *pnt++ = 0x00000000;
  }
}


void  ili9328_fill_full1( uint16_t color )
{
  int  i;
  uint16_t  *dst = (uint16_t *)DEF_LCD_VUR_FB1;

  for( i=0 ; i<DEF_LCD_FB_SIZE/2 ; i++ ) {
    *dst++ = color;
  }
}


void  ili9328_fill_full2( uint16_t color )
{
  int  i;
  uint16_t  *dst = (uint16_t *)DEF_LCD_VUR_FB2;

  for( i=0 ; i<DEF_LCD_FB_SIZE/2 ; i++ ) {
    *dst++ = color;
  }
}


/*
#define  DEF_LCD_VUR_FB1   0xc0040000
#define  DEF_LCD_VUR_FB2   0xc0090000
*/
void  ili9328_copy_bmp_to_v( unsigned char *bmp, unsigned short int *v_top )
{
  unsigned char  *top = bmp + 0x36;
  unsigned char  *rp;
  unsigned short int  rgb;
  int  x, y;

  for( y=0 ; y<240 ; y++ ) {
    rp = top + (240-1-y)*320*3;
    for( x=0 ; x<320 ; x++ ) {
        rgb = ((unsigned short int)(*(rp+2) & 0xf8) << 8)
             |((unsigned short int)(*(rp+1) & 0xfc) << 3)
             |((unsigned short int)(*(rp+0) & 0xf8) >> 3);

        *v_top++ = rgb;
        rp += 3;
    }
  }
}


void  ili9328_copy_fb_to_v1( void )
{
  int  i;
  unsigned long  *src = (unsigned long *)DEF_LCD_FB_TOP;
  unsigned long  *dst = (unsigned long *)DEF_LCD_VUR_FB1;

  for( i=0 ; i<DEF_LCD_FB_SIZE/4 ; i++ ) {
    *dst++ = *src++;
  }
}


void  ili9328_copy_fb_to_v2( void )
{
  int  i;
  unsigned long  *src = (unsigned long *)DEF_LCD_FB_TOP;
  unsigned long  *dst = (unsigned long *)DEF_LCD_VUR_FB2;

  for( i=0 ; i<DEF_LCD_FB_SIZE/4 ; i++ ) {
    *dst++ = *src++;
  }
}


void  ili9328_copy_v_to_r1( void )
{
  uint16_t  x, y;
  uint16_t  *src = (uint16_t *)DEF_LCD_VUR_FB1;
  uint8_t   *dst = (uint8_t *)DEF_LCD_REL_FB1;
  uint8_t   *tmp;

  for( x=0 ; x<320 ; x++ ) {
    tmp = (uint8_t *)(src + (320 * (240-1)) + x);
    for( y=0 ; y<240 ; y++ ) {
        *dst++ = *(tmp+1);
        *dst++ = *tmp;
        tmp -= 320*2;
    }
  }
}


void  ili9328_copy_v_to_r2( void )
{
  uint16_t  x, y;
  uint16_t  *src = (uint16_t *)DEF_LCD_VUR_FB2;
  uint8_t   *dst = (uint8_t *)DEF_LCD_REL_FB2;
  uint8_t   *tmp;

  for( x=0 ; x<320 ; x++ ) {
    tmp = (uint8_t *)(src + (320 * (240-1)) + x);
    for( y=0 ; y<240 ; y++ ) {
        *dst++ = *(tmp+1);
        *dst++ = *tmp;
        tmp -= 320*2;
    }
  }
}


void  ili9328_disp_lcd( void )
{
  uint8_t  dat[4];
  uint8_t  *src = (uint8_t *)DEF_LCD_REL_FB1;

  if( disp_current_lcd == 0 ) {
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 0);   /* /CS:L */
    src = (uint8_t *)DEF_LCD_REL_FB1;
  } else {
    //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 0);   /* /CS2:L */
    GPIOG->BSRR = 0x0080 << 16 ; /* /CS2:L */
    src = (uint8_t *)DEF_LCD_REL_FB2;
  }
  dat[0] = (uint8_t)0x72;
  HAL_SPI_Transmit(&SpiHandle, dat, 1, 1000);

  HAL_SPI_Transmit(&SpiHandle, src, 320*2*80, 1000000);
  src += 320*2*80;
  HAL_SPI_Transmit(&SpiHandle, src, 320*2*80, 1000000);
  src += 320*2*80;
  HAL_SPI_Transmit(&SpiHandle, src, 320*2*80, 1000000);

  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 1);   /* /CS:H */
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 1);   /* /CS2:H */
  GPIOG->BSRR = 0x0080;        /* /CS2:H */
}


void  ili9328_copy_fullx( uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1, uint16_t *src )
{
  uint16_t  x, y;
  uint16_t  *tmp;
  uint8_t  dat[4];

  /* setblock() */
  writecom16(0x0050); 
  writedat16(y0); 
  writecom16(0x0051); 
  writedat16(y1); 
  writecom16(0x0052); 
  writedat16(x0); 
  writecom16(0x0053); 
  writedat16(x1); 
  writecom16(0x0020); 
  writedat16(x0); 
  writecom16(0x0021); 
  writedat16(y0); 
  writecom16(0x0022); 

 if( disp_current_lcd == 0 ) {
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 0);   /* /CS:L */
 } else {
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 0);   /* /CS2:L */
  GPIOG->BSRR = 0x0080 << 16 ; /* /CS2:L */
 }
  dat[0] = (uint8_t)0x72;
  HAL_SPI_Transmit(&SpiHandle, dat, 1, 1000);

  for( x=x0 ; x<x1 ; x++ ) {
    tmp = src + (320 * (240-1)) + x;
    for( y=y0 ; y<y1 ; y++ ) {
        dat[0] = (uint8_t)((*tmp >> 8) & 0xff); 
        dat[1] = (uint8_t)(*tmp & 0xff); 
        HAL_SPI_Transmit(&SpiHandle, dat, 2, 1000);
        tmp -= 320;
    }
  }

  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 1);   /* /CS:H */
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 1);   /* /CS2:H */
  GPIOG->BSRR = 0x0080;        /* /CS2:H */
}


void  ili9328_copy_fullx21( uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1, uint16_t *src )
{
  uint16_t  x, y;
  uint16_t  *tmp;
  uint8_t  dat[4];

  /* setblock() */
  writecom16(0x0050); 
  writedat16(y0); 
  writecom16(0x0051); 
  writedat16(y1); 
  writecom16(0x0052); 
  writedat16(x0); 
  writecom16(0x0053); 
  writedat16(x1); 
  writecom16(0x0020); 
  writedat16(x0); 
  writecom16(0x0021); 
  writedat16(y0); 
  writecom16(0x0022); 

 if( disp_current_lcd == 0 ) {
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, 0);   /* /CS:L */
 } else {
  //HAL_GPIO_WritePin(GPIOG, GPIO_PIN_7, 0);   /* /CS2:L */
  GPIOG->BSRR = 0x0080 << 16 ; /* /CS2:L */
 }
  dat[0] = (uint8_t)0x72;
  HAL_SPI_Transmit(&SpiHandle, dat, 1, 1000);

  for( x=x0 ; x<x1 ; x++ ) {
    tmp = src + (320 * (240-1)) + x;
    for( y=y0 ; y<y1 ; y++ ) {
        dat[0] = (uint8_t)((*tmp >> 8) & 0xff); 
        dat[1] = (uint8_t)(*tmp & 0xff); 
        HAL_SPI_Transmit(&SpiHandle, dat, 2, 1000);
        tmp -= 320;
    }
  }
}


void  ili9328_copy_fullx22( uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1, uint16_t *src )
{
  uint8_t  *dat = (uint8_t *)src;
  HAL_SPI_Transmit(&SpiHandle, dat, 320*2*80, 1000000);
  dat += 320*2*80;
  HAL_SPI_Transmit(&SpiHandle, dat, 320*2*80, 1000000);
  dat += 320*2*80;
  HAL_SPI_Transmit(&SpiHandle, dat, 320*2*80, 1000000);
}


/***********************************************************************/


/***********************************************************************/
void  xxHAL_SDRAM_MspInit( void )
{
  GPIO_InitTypeDef  GPIO_Init_Structure;

  /* Enable GPIO clocks */
  __HAL_RCC_GPIOC_CLK_ENABLE();
  __HAL_RCC_GPIOD_CLK_ENABLE();
  __HAL_RCC_GPIOE_CLK_ENABLE();
  __HAL_RCC_GPIOF_CLK_ENABLE();
  __HAL_RCC_GPIOG_CLK_ENABLE();
  __HAL_RCC_GPIOH_CLK_ENABLE();

  /* Enable FMC clock */
  __HAL_RCC_FMC_CLK_ENABLE();

  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Pull      = GPIO_PULLUP;
  GPIO_Init_Structure.Speed     = GPIO_SPEED_FAST;
  GPIO_Init_Structure.Alternate = GPIO_AF12_FMC;


  /* GPIOC configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_3;
  HAL_GPIO_Init(GPIOC, &GPIO_Init_Structure);
  
  /* GPIOD configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_8 | GPIO_PIN_9 |
                              GPIO_PIN_10 | GPIO_PIN_14 | GPIO_PIN_15;
  HAL_GPIO_Init(GPIOD, &GPIO_Init_Structure);

  /* GPIOE configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_7| GPIO_PIN_8 | GPIO_PIN_9       |\
                              GPIO_PIN_10 | GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13 | GPIO_PIN_14 |\
                              GPIO_PIN_15;
  HAL_GPIO_Init(GPIOE, &GPIO_Init_Structure);
  
  /* GPIOF configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2| GPIO_PIN_3 | GPIO_PIN_4      |\
                              GPIO_PIN_5 | GPIO_PIN_11 | GPIO_PIN_12 | GPIO_PIN_13 | GPIO_PIN_14 |\
                              GPIO_PIN_15;
  HAL_GPIO_Init(GPIOF, &GPIO_Init_Structure);
  
  /* GPIOG configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_4| GPIO_PIN_5 | GPIO_PIN_8 |\
                              GPIO_PIN_15;
  HAL_GPIO_Init(GPIOG, &GPIO_Init_Structure);

  /* GPIOH configuration */
  GPIO_Init_Structure.Pin   = GPIO_PIN_3 | GPIO_PIN_5;
  HAL_GPIO_Init(GPIOH, &GPIO_Init_Structure);
}


static  void  SDRAM_Config( void )
{
  int  chk ;

  syslog(3, "SDRAM check Start.");
  HAL_Delay( 100 );

  /*##-1- Configure the SDRAM device #########################################*/
  /* SDRAM device configuration */
  hsdram.Instance = FMC_SDRAM_DEVICE;

  SDRAM_Timing.LoadToActiveDelay    = 2;
  SDRAM_Timing.ExitSelfRefreshDelay = 6;
  SDRAM_Timing.SelfRefreshTime      = 4;
  SDRAM_Timing.RowCycleDelay        = 6;
  SDRAM_Timing.WriteRecoveryTime    = 2;
  SDRAM_Timing.RPDelay              = 2;
  SDRAM_Timing.RCDDelay             = 2;

  hsdram.Init.SDBank             = FMC_SDRAM_BANK1;
  hsdram.Init.ColumnBitsNumber   = FMC_SDRAM_COLUMN_BITS_NUM_8;
  hsdram.Init.RowBitsNumber      = FMC_SDRAM_ROW_BITS_NUM_12;
  hsdram.Init.MemoryDataWidth    = SDRAM_MEMORY_WIDTH;
  hsdram.Init.InternalBankNumber = FMC_SDRAM_INTERN_BANKS_NUM_4;
  hsdram.Init.CASLatency         = FMC_SDRAM_CAS_LATENCY_2;
  hsdram.Init.WriteProtection    = FMC_SDRAM_WRITE_PROTECTION_DISABLE;
  hsdram.Init.SDClockPeriod      = SDCLOCK_PERIOD;
  hsdram.Init.ReadBurst          = FMC_SDRAM_RBURST_ENABLE;
  hsdram.Init.ReadPipeDelay      = FMC_SDRAM_RPIPE_DELAY_0;

  /* Initialize the SDRAM controller */
  if(HAL_SDRAM_Init(&hsdram, &SDRAM_Timing) != HAL_OK)
  {
    /* Initialization Error */
    Error_Handler();
  }

  /* Program the SDRAM external device */
  BSP_SDRAM_Initialization_Sequence(&hsdram, &command);

  /* Fill the buffer to write */
  Fill_Buffer(aTxBuffer, BUFFER_SIZE, 0xA244250F);

  /* Write data to the SDRAM memory */
  for (uwIndex = 0; uwIndex < BUFFER_SIZE; uwIndex++)
  {
    *(__IO uint32_t*) (SDRAM_BANK_ADDR + WRITE_READ_ADDR + 4*uwIndex) = aTxBuffer[uwIndex];
  }

  /* Read back data from the SDRAM memory */
  for (uwIndex = 0; uwIndex < BUFFER_SIZE; uwIndex++)
  {
    aRxBuffer[uwIndex] = *(__IO uint32_t*) (SDRAM_BANK_ADDR + WRITE_READ_ADDR + 4*uwIndex);
  }

  /*##-3- Checking data integrity ############################################*/
  chk = 0 ;
  for (uwIndex = 0; (uwIndex < BUFFER_SIZE) && (uwWriteReadStatus == 0); uwIndex++)
  {
    chk++ ;
    if (aRxBuffer[uwIndex] != aTxBuffer[uwIndex])
    {
      uwWriteReadStatus++;
    }
  }

  if (uwWriteReadStatus != PASSED)
  {
    syslog(3, "SDRAM check NG.  chk=%d  uwWriteReadStatus=%d", chk, uwWriteReadStatus);
    while(1)
    {
      /* KO, Toggle LED1 with a period of 200 ms */
      BSP_LED_Toggle(LED1);
      HAL_Delay(200);
    }
  }
  else
  {
    syslog(3, "SDRAM check OK.");
  }
}


void xxHAL_LTDC_MspInit( void )
{  
  GPIO_InitTypeDef GPIO_Init_Structure;
  
  /* Enable the LTDC Clock */
  __HAL_RCC_LTDC_CLK_ENABLE();
  
  /******************** LTDC Pins configuration *************************/
  /* Enable GPIOs clock */
  __HAL_RCC_GPIOE_CLK_ENABLE();
  __HAL_RCC_GPIOG_CLK_ENABLE();
  __HAL_RCC_GPIOI_CLK_ENABLE();
  __HAL_RCC_GPIOJ_CLK_ENABLE();
  __HAL_RCC_GPIOK_CLK_ENABLE();
  
  /*** LTDC Pins configuration ***/
  /* GPIOE configuration */
  GPIO_Init_Structure.Pin       = GPIO_PIN_4;
  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Pull      = GPIO_NOPULL;
  GPIO_Init_Structure.Speed     = GPIO_SPEED_FAST;
  GPIO_Init_Structure.Alternate = GPIO_AF14_LTDC;  
  HAL_GPIO_Init(GPIOE, &GPIO_Init_Structure);

  /* GPIOG configuration */
  GPIO_Init_Structure.Pin       = GPIO_PIN_12;
  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Alternate = GPIO_AF9_LTDC;
  HAL_GPIO_Init(GPIOG, &GPIO_Init_Structure);

  /* GPIOI LTDC alternate configuration */
  GPIO_Init_Structure.Pin       = GPIO_PIN_9 | GPIO_PIN_10 | \
                                  GPIO_PIN_14 | GPIO_PIN_15;
  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Alternate = GPIO_AF14_LTDC;
  HAL_GPIO_Init(GPIOI, &GPIO_Init_Structure);

  /* GPIOJ configuration */  
  GPIO_Init_Structure.Pin       = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_3 | \
                                  GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7 | \
                                  GPIO_PIN_8 | GPIO_PIN_9 | GPIO_PIN_10 | GPIO_PIN_11 | \
                                  GPIO_PIN_13 | GPIO_PIN_14 | GPIO_PIN_15;
  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Alternate = GPIO_AF14_LTDC;
  HAL_GPIO_Init(GPIOJ, &GPIO_Init_Structure);  

  /* GPIOK configuration */  
  GPIO_Init_Structure.Pin       = GPIO_PIN_0 | GPIO_PIN_1 | GPIO_PIN_2 | GPIO_PIN_4 | \
                                  GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7;
  GPIO_Init_Structure.Mode      = GPIO_MODE_AF_PP;
  GPIO_Init_Structure.Alternate = GPIO_AF14_LTDC;
  HAL_GPIO_Init(GPIOK, &GPIO_Init_Structure);
  
  /* LCD_DISP GPIO configuration */
  GPIO_Init_Structure.Pin       = GPIO_PIN_12;     /* LCD_DISP pin has to be manually controlled */
  GPIO_Init_Structure.Mode      = GPIO_MODE_OUTPUT_PP;
  HAL_GPIO_Init(GPIOI, &GPIO_Init_Structure);

  /* LCD_BL_CTRL GPIO configuration */
  GPIO_Init_Structure.Pin       = GPIO_PIN_3;  /* LCD_BL_CTRL pin has to be manually controlled */
  GPIO_Init_Structure.Mode      = GPIO_MODE_OUTPUT_PP;
  HAL_GPIO_Init(GPIOK, &GPIO_Init_Structure);
  
  /* Assert display enable LCD_DISP pin */
  HAL_GPIO_WritePin(LCD_DISP_GPIO_PORT, LCD_DISP_PIN, GPIO_PIN_SET);

  /* Assert backlight LCD_BL_CTRL pin */
  HAL_GPIO_WritePin(LCD_BL_CTRL_GPIO_PORT, LCD_BL_CTRL_PIN, GPIO_PIN_SET);

  {
    RCC_PeriphCLKInitTypeDef  PeriphClkInitStruct;

    /* LCD clock configuration */
    /* PLLSAI_VCO Input = HSE_VALUE/PLL_M = 1 Mhz */
    /* PLLSAI_VCO Output = PLLSAI_VCO Input * PLLSAIN = 192 Mhz */
    /* PLLLCDCLK = PLLSAI_VCO Output/PLLSAIR = 192/5 = 38.4 Mhz */
    /* LTDC clock frequency = PLLLCDCLK / LTDC_PLLSAI_DIVR_4 = 38.4/4 = 9.6Mhz */
    PeriphClkInitStruct.PeriphClockSelection = RCC_PERIPHCLK_LTDC;
    PeriphClkInitStruct.PLLSAI.PLLSAIN = 192;
    PeriphClkInitStruct.PLLSAI.PLLSAIR = 5;
    PeriphClkInitStruct.PLLSAIDivR = RCC_PLLSAIDIVR_4;
    HAL_RCCEx_PeriphCLKConfig(&PeriphClkInitStruct);  
  }
}


LTDC_HandleTypeDef hltdc_F;
LTDC_LayerCfgTypeDef      pLayerCfg;

static void LCD_Config(void)
{ 
  /* LTDC Initialization -------------------------------------------------------*/
  
  /* Polarity configuration */
  /* Initialize the horizontal synchronization polarity as active low */
  hltdc_F.Init.HSPolarity = LTDC_HSPOLARITY_AL;
  /* Initialize the vertical synchronization polarity as active low */ 
  hltdc_F.Init.VSPolarity = LTDC_VSPOLARITY_AL; 
  /* Initialize the data enable polarity as active low */ 
  hltdc_F.Init.DEPolarity = LTDC_DEPOLARITY_AL; 
  /* Initialize the pixel clock polarity as input pixel clock */  
  hltdc_F.Init.PCPolarity = LTDC_PCPOLARITY_IPC;
  
  /* The RK043FN48H LCD 480x272 is selected */
  /* Timing Configuration */
  hltdc_F.Init.HorizontalSync = (RK043FN48H_HSYNC - 1);
  hltdc_F.Init.VerticalSync = (RK043FN48H_VSYNC - 1);
  hltdc_F.Init.AccumulatedHBP = (RK043FN48H_HSYNC + RK043FN48H_HBP - 1);
  hltdc_F.Init.AccumulatedVBP = (RK043FN48H_VSYNC + RK043FN48H_VBP - 1);
  hltdc_F.Init.AccumulatedActiveH = (RK043FN48H_HEIGHT + RK043FN48H_VSYNC + RK043FN48H_VBP - 1);
  hltdc_F.Init.AccumulatedActiveW = (RK043FN48H_WIDTH + RK043FN48H_HSYNC + RK043FN48H_HBP - 1);
  hltdc_F.Init.TotalHeigh = (RK043FN48H_HEIGHT + RK043FN48H_VSYNC + RK043FN48H_VBP + RK043FN48H_VFP - 1);
  hltdc_F.Init.TotalWidth = (RK043FN48H_WIDTH + RK043FN48H_HSYNC + RK043FN48H_HBP + RK043FN48H_HFP - 1);
  
  /* Configure R,G,B component values for LCD background color : all black background */
  hltdc_F.Init.Backcolor.Blue = 0;
  hltdc_F.Init.Backcolor.Green = 0;
  hltdc_F.Init.Backcolor.Red = 0;

  hltdc_F.Instance = LTDC;
  
/* Layer1 Configuration ------------------------------------------------------*/
  
  /* Windowing configuration */ 
  /* In this case all the active display area is used to display a picture then :
     Horizontal start = horizontal synchronization + Horizontal back porch = 43 
     Vertical start   = vertical synchronization + vertical back porch     = 12
     Horizontal stop = Horizontal start + window width -1 = 43 + 480 -1 
     Vertical stop   = Vertical start + window height -1  = 12 + 272 -1      */
  pLayerCfg.WindowX0 = 0;
  pLayerCfg.WindowX1 = 480;
  pLayerCfg.WindowY0 = 0;
  pLayerCfg.WindowY1 = 272;
  
  /* Pixel Format configuration*/ 
  pLayerCfg.PixelFormat = LTDC_PIXEL_FORMAT_RGB565;
  
  /* Start Address configuration : frame buffer is located at FLASH memory */
  pLayerCfg.FBStartAdress = (uint32_t)0xc0000000 ;
  
  /* Alpha constant (255 == totally opaque) */
  pLayerCfg.Alpha = 255;
  
  /* Default Color configuration (configure A,R,G,B component values) : no background color */
  pLayerCfg.Alpha0 = 0; /* fully transparent */
  pLayerCfg.Backcolor.Blue = 0;
  pLayerCfg.Backcolor.Green = 0;
  pLayerCfg.Backcolor.Red = 0;
  
  /* Configure blending factors */
  pLayerCfg.BlendingFactor1 = LTDC_BLENDING_FACTOR1_CA;
  pLayerCfg.BlendingFactor2 = LTDC_BLENDING_FACTOR2_CA;
  
  /* Configure the number of lines and number of pixels per line */
  pLayerCfg.ImageWidth  = 480;
  pLayerCfg.ImageHeight = 272;
  
  /* Configure the LTDC */  
  if(HAL_LTDC_Init(&hltdc_F) != HAL_OK)
  {
    /* Initialization Error */
    Error_Handler(); 
  }
    
  /* Configure the Layer*/
  if(HAL_LTDC_ConfigLayer(&hltdc_F, &pLayerCfg, 1) != HAL_OK)
  {
    /* Initialization Error */
    Error_Handler(); 
  }  
}


/*************** for CAMARA **********************/
#include "ov9655.h"
#define CAMERA_VGA_RES_X          640
#define CAMERA_VGA_RES_Y          480
#define CAMERA_480x272_RES_X      480
#define CAMERA_480x272_RES_Y      272
#define CAMERA_QVGA_RES_X         320
#define CAMERA_QVGA_RES_Y         240
#define CAMERA_QQVGA_RES_X        160
#define CAMERA_QQVGA_RES_Y        120

typedef enum
{
  CAMERA_OK            = 0x00,
  CAMERA_ERROR         = 0x01,
  CAMERA_TIMEOUT       = 0x02,
  CAMERA_NOT_DETECTED  = 0x03,
  CAMERA_NOT_SUPPORTED = 0x04

} Camera_StatusTypeDef;

#define RESOLUTION_R160x120      CAMERA_R160x120      /* QQVGA Resolution     */
#define RESOLUTION_R320x240      CAMERA_R320x240      /* QVGA Resolution      */
#define RESOLUTION_R480x272      CAMERA_R480x272      /* 480x272 Resolution   */
#define RESOLUTION_R640x480      CAMERA_R640x480      /* VGA Resolution       */

DCMI_HandleTypeDef  hDcmiHandler;
CAMERA_DrvTypeDef   *camera_drv;
/* Camera current resolution naming (QQVGA, VGA, ...) */
static uint32_t CameraCurrentResolution;
/* Camera module I2C HW address */
static uint32_t CameraHwAddress;


void BSP_CAMERA_MspInit(DCMI_HandleTypeDef *hdcmi, void *Params)
{
  static DMA_HandleTypeDef hdma_handler;
  GPIO_InitTypeDef gpio_init_structure;
  
  /*** Enable peripherals and GPIO clocks ***/
  /* Enable DCMI clock */
  __HAL_RCC_DCMI_CLK_ENABLE();

  /* Enable DMA2 clock */
  __HAL_RCC_DMA2_CLK_ENABLE();
  
  /* Enable GPIO clocks */
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOD_CLK_ENABLE();
  __HAL_RCC_GPIOE_CLK_ENABLE();
  __HAL_RCC_GPIOG_CLK_ENABLE();
  __HAL_RCC_GPIOH_CLK_ENABLE();

  /*** Configure the GPIO ***/
  /* Configure DCMI GPIO as alternate function */
  gpio_init_structure.Pin       = GPIO_PIN_4 | GPIO_PIN_6;
  gpio_init_structure.Mode      = GPIO_MODE_AF_PP;
  gpio_init_structure.Pull      = GPIO_PULLUP;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  gpio_init_structure.Alternate = GPIO_AF13_DCMI;
  HAL_GPIO_Init(GPIOA, &gpio_init_structure);

  gpio_init_structure.Pin       = GPIO_PIN_3;
  gpio_init_structure.Mode      = GPIO_MODE_AF_PP;
  gpio_init_structure.Pull      = GPIO_PULLUP;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  gpio_init_structure.Alternate = GPIO_AF13_DCMI;
  HAL_GPIO_Init(GPIOD, &gpio_init_structure);

  gpio_init_structure.Pin       = GPIO_PIN_5 | GPIO_PIN_6;
  gpio_init_structure.Mode      = GPIO_MODE_AF_PP;
  gpio_init_structure.Pull      = GPIO_PULLUP;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  gpio_init_structure.Alternate = GPIO_AF13_DCMI;
  HAL_GPIO_Init(GPIOE, &gpio_init_structure);

  gpio_init_structure.Pin       = GPIO_PIN_9;
  gpio_init_structure.Mode      = GPIO_MODE_AF_PP;
  gpio_init_structure.Pull      = GPIO_PULLUP;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  gpio_init_structure.Alternate = GPIO_AF13_DCMI;
  HAL_GPIO_Init(GPIOG, &gpio_init_structure);

  gpio_init_structure.Pin       = GPIO_PIN_9 | GPIO_PIN_10  | GPIO_PIN_11  |\
                                  GPIO_PIN_12 | GPIO_PIN_14;
  gpio_init_structure.Mode      = GPIO_MODE_AF_PP;
  gpio_init_structure.Pull      = GPIO_PULLUP;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  gpio_init_structure.Alternate = GPIO_AF13_DCMI;
  HAL_GPIO_Init(GPIOH, &gpio_init_structure);

  /*** Configure the DMA ***/
  /* Set the parameters to be configured */
  hdma_handler.Init.Channel             = DMA_CHANNEL_1;
  hdma_handler.Init.Direction           = DMA_PERIPH_TO_MEMORY;
  hdma_handler.Init.PeriphInc           = DMA_PINC_DISABLE;
  hdma_handler.Init.MemInc              = DMA_MINC_ENABLE;
  hdma_handler.Init.PeriphDataAlignment = DMA_PDATAALIGN_WORD;
  hdma_handler.Init.MemDataAlignment    = DMA_MDATAALIGN_WORD;
  hdma_handler.Init.Mode                = DMA_CIRCULAR;
  hdma_handler.Init.Priority            = DMA_PRIORITY_HIGH;
  hdma_handler.Init.FIFOMode            = DMA_FIFOMODE_DISABLE;
  hdma_handler.Init.FIFOThreshold       = DMA_FIFO_THRESHOLD_FULL;
  hdma_handler.Init.MemBurst            = DMA_MBURST_SINGLE;
  hdma_handler.Init.PeriphBurst         = DMA_PBURST_SINGLE; 

  hdma_handler.Instance = DMA2_Stream1;

  /* Associate the initialized DMA handle to the DCMI handle */
  __HAL_LINKDMA(hdcmi, DMA_Handle, hdma_handler);
  
  /* Configure the DMA stream */
  HAL_DMA_Init(hdcmi->DMA_Handle);  
}


static uint32_t GetSize(uint32_t resolution)
{ 
  uint32_t size = 0;
  
  /* Get capture size */
  switch (resolution)
  {
  case CAMERA_R160x120:
    {
      size =  0x2580;
    }
    break;    
  case CAMERA_R320x240:
    {
      size =  0x9600;
    }
    break;
  case CAMERA_R480x272:
    {
      size =  0xFF00;
    }
    break;
  case CAMERA_R640x480:
    {
      size =  0x25800;
    }    
    break;
  default:
    {
      break;
    }
  }
  
  return size;
}


/**
  * @brief  Starts the camera capture in continuous mode.
  * @param  buff: pointer to the camera output buffer
  * @retval None
  */
void BSP_CAMERA_ContinuousStart(uint8_t *buff)
{ 
  /* Start the camera capture */
  HAL_DCMI_Start_DMA(&hDcmiHandler, DCMI_MODE_CONTINUOUS, (uint32_t)buff, GetSize(CameraCurrentResolution));
}

/**
  * @brief  Starts the camera capture in snapshot mode.
  * @param  buff: pointer to the camera output buffer
  * @retval None
  */
void BSP_CAMERA_SnapshotStart(uint8_t *buff)
{ 
  /* Start the camera capture */
  HAL_DCMI_Start_DMA(&hDcmiHandler, DCMI_MODE_SNAPSHOT, (uint32_t)buff, GetSize(CameraCurrentResolution));
}

/**
  * @brief Suspend the CAMERA capture 
  * @retval None
  */
void BSP_CAMERA_Suspend(void) 
{
  /* Suspend the Camera Capture */
  HAL_DCMI_Suspend(&hDcmiHandler);  
}

/**
  * @brief Resume the CAMERA capture 
  * @retval None
  */
void BSP_CAMERA_Resume(void) 
{
  /* Start the Camera Capture */
  HAL_DCMI_Resume(&hDcmiHandler);
}


void BSP_CAMERA_PwrDown(void)
{
  GPIO_InitTypeDef gpio_init_structure;

  /* Enable GPIO clock */
  __HAL_RCC_GPIOH_CLK_ENABLE();

  /*** Configure the GPIO ***/
  /* Configure DCMI GPIO as alternate function */
  gpio_init_structure.Pin       = GPIO_PIN_13;
  gpio_init_structure.Mode      = GPIO_MODE_OUTPUT_PP;
  gpio_init_structure.Pull      = GPIO_NOPULL;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  HAL_GPIO_Init(GPIOH, &gpio_init_structure);

  /* Assert the camera POWER_DOWN pin (active high) */
  HAL_GPIO_WritePin(GPIOH, GPIO_PIN_13, GPIO_PIN_SET);
}


/**
  * @brief  Stop the CAMERA capture 
  * @retval Camera status
  */
uint8_t BSP_CAMERA_Stop(void) 
{
  uint8_t status = CAMERA_ERROR;

  if(HAL_DCMI_Stop(&hDcmiHandler) == HAL_OK)
  {
     status = CAMERA_OK;
  }
  
  /* Set Camera in Power Down */
  BSP_CAMERA_PwrDown();

  return status;
}


void BSP_CAMERA_PwrUp(void)
{
  GPIO_InitTypeDef gpio_init_structure;

  /* Enable GPIO clock */
  __HAL_RCC_GPIOH_CLK_ENABLE();

  /*** Configure the GPIO ***/
  /* Configure DCMI GPIO as alternate function */
  gpio_init_structure.Pin       = GPIO_PIN_13;
  gpio_init_structure.Mode      = GPIO_MODE_OUTPUT_PP;
  gpio_init_structure.Pull      = GPIO_NOPULL;
  gpio_init_structure.Speed     = GPIO_SPEED_HIGH;
  HAL_GPIO_Init(GPIOH, &gpio_init_structure);

  /* De-assert the camera POWER_DOWN pin (active high) */
  HAL_GPIO_WritePin(GPIOH, GPIO_PIN_13, GPIO_PIN_RESET);

  HAL_Delay(3);     /* POWER_DOWN de-asserted during 3ms */
}


uint8_t BSP_CAMERA_Init(uint32_t Resolution)
{ 
  DCMI_HandleTypeDef *phdcmi;
  uint8_t status = CAMERA_ERROR;

  /* Get the DCMI handle structure */
  phdcmi = &hDcmiHandler;

  /*** Configures the DCMI to interface with the camera module ***/
  /* DCMI configuration */
  phdcmi->Init.CaptureRate      = DCMI_CR_ALL_FRAME;
  phdcmi->Init.HSPolarity       = DCMI_HSPOLARITY_LOW;
  //phdcmi->Init.HSPolarity       = DCMI_HSPOLARITY_HIGH;
  phdcmi->Init.SynchroMode      = DCMI_SYNCHRO_HARDWARE;
  phdcmi->Init.VSPolarity       = DCMI_VSPOLARITY_HIGH;
  //phdcmi->Init.VSPolarity       = DCMI_VSPOLARITY_LOW;
  phdcmi->Init.ExtendedDataMode = DCMI_EXTEND_DATA_8B;
  phdcmi->Init.PCKPolarity      = DCMI_PCKPOLARITY_RISING;
  //phdcmi->Init.PCKPolarity      = DCMI_PCKPOLARITY_FALLING;
  phdcmi->Instance              = DCMI;

  /* Power up camera */
  BSP_CAMERA_PwrUp();

  /* Read ID of Camera module via I2C */
#if 1
  syslog( 3, "ov9655_ReadID(CAMERA_I2C_ADDRESS)=0x%04x", ov9655_ReadID(CAMERA_I2C_ADDRESS));
  if(ov9655_ReadID(CAMERA_I2C_ADDRESS) == OV9655_ID)
#else
  if(1)
#endif
  {
    /* Initialize the camera driver structure */
    camera_drv = &ov9655_drv;
    CameraHwAddress = CAMERA_I2C_ADDRESS;

    /* DCMI Initialization */
    BSP_CAMERA_MspInit(&hDcmiHandler, NULL);
    HAL_DCMI_Init(phdcmi);

    /* Camera Module Initialization via I2C to the wanted 'Resolution' */
    if (Resolution == CAMERA_R480x272)
    {     /* For 480x272 resolution, the OV9655 sensor is set to VGA resolution
           * as OV9655 doesn't supports 480x272 resolution,
           * then DCMI is configured to output a 480x272 cropped window */
      camera_drv->Init(CameraHwAddress, CAMERA_R640x480);

      if( imgpro_gra_size == 1 ) {
        /** 480x272 **/
       HAL_DCMI_ConfigCROP(phdcmi,           /* Crop in the middle of the VGA picture */
                         (CAMERA_VGA_RES_X - CAMERA_480x272_RES_X)/2,
                         (CAMERA_VGA_RES_Y - CAMERA_480x272_RES_Y)/2,
                         (CAMERA_480x272_RES_X * 2) - 1,
                          CAMERA_480x272_RES_Y - 1);
      } else {
        /* 480x320 */
        HAL_DCMI_ConfigCROP(phdcmi,           /* Crop in the middle of the VGA picture */
                         (CAMERA_VGA_RES_X - 480)/2,
                         (CAMERA_VGA_RES_Y - 320)/2,
                         (480 * 2) - 1,
                          320 - 1);
      }
      HAL_DCMI_EnableCROP(phdcmi);
    }
    else
    {
      camera_drv->Init(CameraHwAddress, Resolution);
      HAL_DCMI_DisableCROP(phdcmi);
    }

    CameraCurrentResolution = Resolution;

    /* Return CAMERA_OK status */
    status = CAMERA_OK;
  }
  else
  {
    /* Return CAMERA_NOT_SUPPORTED status */
    status = CAMERA_NOT_SUPPORTED;
    syslog(3, "CAMERA_NOT_SUPPORTED");
  }

  return status;
}


int  cam_button_flag ;



/****************************************
   Device initialize
****************************************/
void  imgpro_device_init(void)
{
  CPU_CACHE_Enable();

  //BSP_LED_Init(LED1);
  BSP_PB_Init(BUTTON_KEY, BUTTON_MODE_GPIO);
  usrbtn_init_PB8();
  usrbtn_init_2();
  usrbtn_init_3();

  xxHAL_UART_MspInit();
  UART_Config();

  syslog(3, "xxHAL_SDRAM_MspInit()");
  xxHAL_SDRAM_MspInit();
  SDRAM_Config() ;

  xxHAL_SPI_MspInit();
  SPI_Config();

  imgpro_sdcard_init();

  disp_current_lcd = 0;
  ili9328_clr_framebuffer();

  xxHAL_LTDC_MspInit();
  LCD_Config() ;

  if( imgpro_gra_size == 1 ) {
    /** 480x320 **/
    BSP_CAMERA_Init( CAMERA_R480x272 );

    pLayerCfg.ImageWidth  = 480;
    pLayerCfg.ImageHeight = 272;
    pLayerCfg.WindowX0 = 0;
    pLayerCfg.WindowX1 = 480;
    pLayerCfg.WindowY0 = 0;
    pLayerCfg.WindowY1 = 272;
    pLayerCfg.FBStartAdress = (uint32_t)0xc0000000 ;

  } else if( imgpro_gra_size == 2 ) {
    /** 640x480 **/
    BSP_CAMERA_Init( CAMERA_R640x480 );

    pLayerCfg.ImageWidth  = 640;
    pLayerCfg.ImageHeight = 480;
    pLayerCfg.WindowX0 = 0;
    pLayerCfg.WindowX1 = 480;
    pLayerCfg.WindowY0 = 0;
    pLayerCfg.WindowY1 = 272;
    pLayerCfg.FBStartAdress = (uint32_t)0xc0000000 ;

  } else {
    /** 320x240 **/
    BSP_CAMERA_Init( CAMERA_R320x240 );

    pLayerCfg.ImageWidth  = 320;
    pLayerCfg.ImageHeight = 240;
    pLayerCfg.WindowX0 = 80;
    pLayerCfg.WindowX1 = 400;
    pLayerCfg.WindowY0 = 16;
    pLayerCfg.WindowY1 = 256;
    pLayerCfg.FBStartAdress = (uint32_t)0xc0000000 ;
  }

    HAL_LTDC_ConfigLayer(&hltdc_F, &pLayerCfg, 1) ;

  syslog(3, "ili9328_init()");
  disp_current_lcd = 0;
  ili9328_init();
  ili9328_setblock( 0, 320, 0, 240 );
  disp_current_lcd = 1;
  ili9328_init();
  ili9328_setblock( 0, 320, 0, 240 );

  disp_current_lcd = 0;
  ili9328_copy_bmp_to_v( (unsigned char *)bg320x240_T2_bmp, (unsigned short int *)DEF_LCD_VUR_FB1 );
  ili9328_copy_v_to_r1();
  ili9328_disp_lcd();

  disp_current_lcd = 1;
  ili9328_copy_bmp_to_v( (unsigned char *)bg320x240_T1_bmp, (unsigned short int *)DEF_LCD_VUR_FB2 );
  ili9328_copy_v_to_r2();
  ili9328_disp_lcd();
}


int  cam_main_event ;
int  cam_key_event ;


/****************************************
    key scan function
****************************************/
void  keytask( void )
{
    cam_button_flag = 0 ;
    cam_key_event = 0x0000 ;

    while( 1 ) {

        if( cam_button_flag == 0 ) {
            push_button_kind = 0;
            if( (BSP_PB_GetState(BUTTON_KEY) != RESET) || usrbtn_getbtn_PB8() == RESET ) {
                push_button_kind = 1;
            } else if( usrbtn_getbtn_2() == RESET ) {
                push_button_kind = 2;
            } else if( usrbtn_getbtn_3() == RESET ) {
                push_button_kind = 3;
            }

            if( push_button_kind != 0 ) {
                cam_button_flag = 1;
                cam_main_event |= 0x0001;
            }
        }

        if( (cam_button_flag == 1) && (cam_key_event & 0x0001)
            && ((BSP_PB_GetState(BUTTON_KEY) == RESET) && (usrbtn_getbtn_PB8() != RESET)
                && (usrbtn_getbtn_2() != RESET) && (usrbtn_getbtn_3() != RESET) ) ) {
            cam_button_flag = 0;
            cam_key_event &= ~0x0001;
        }

        tslp_tsk( 10 );
    }
}


/****************************************
    flash function
****************************************/
void  imgpro_flash( void )
{
    /* dummy funciton */

    if( cam_main_event & 0x0002 ) {
        /* flash is busy */
        return ;
    }

    cam_main_event |=  0x0002 ;
    cam_main_event &= ~0x0001;
}


/****************************************
    main function
****************************************/
void  imgpro_main( void )
{
    syslog(3, "START imgpro_main()");

    cam_main_event = 0x0000 ;

    /* call USER initialize function */
    userfunc_init();

    /* start camera */
    BSP_CAMERA_ContinuousStart( (void *)0xc0000000 );

    /* check and action EVENT */
    while (1) {
        if( cam_main_event & 0x0001 ) {
            imgpro_flash();
        }

        if( cam_main_event & 0x0002 ) {
            /* call USER main function */
            userfunc_main( push_button_kind );
        }

        if( cam_main_event & 0x0004 ) {
            /* request scan-restart to key task */
            cam_key_event |= 0x0001;
        }

        tslp_tsk( 10 );
    }
}


/**
  * @brief  This function is executed in case of error occurrence.
  * @param  None
  * @retval None
  */
static void Error_Handler(void)
{
  /* User may add here some code to deal with this error */
  while (1)
  {
    /* Toggle LED1 with a period of 1 s */
    BSP_LED_Toggle(LED1);
    HAL_Delay(1000);
  }
}

/**
  * @brief  Perform the SDRAM external memory initialization sequence
  * @param  hsdram: SDRAM handle
  * @param  Command: Pointer to SDRAM command structure
  * @retval None
  */
static void BSP_SDRAM_Initialization_Sequence(SDRAM_HandleTypeDef *hsdram, FMC_SDRAM_CommandTypeDef *Command)
{
  __IO uint32_t tmpmrd =0;
  /* Step 3:  Configure a clock configuration enable command */
  Command->CommandMode = FMC_SDRAM_CMD_CLK_ENABLE;
  Command->CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
  Command->AutoRefreshNumber = 1;
  Command->ModeRegisterDefinition = 0;

  /* Send the command */
  HAL_SDRAM_SendCommand(hsdram, Command, SDRAM_TIMEOUT);

  syslog(3, "BSP_SDRAM_Initialization_Sequence()-1");

  /* Step 4: Insert 100 us minimum delay */
  /* Inserted delay is equal to 1 ms due to systick time base unit (ms) */
  HAL_Delay(1);

  /* Step 5: Configure a PALL (precharge all) command */
  Command->CommandMode = FMC_SDRAM_CMD_PALL;
  Command->CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
  Command->AutoRefreshNumber = 1;
  Command->ModeRegisterDefinition = 0;

  /* Send the command */
  HAL_SDRAM_SendCommand(hsdram, Command, SDRAM_TIMEOUT);

  /* Step 6 : Configure a Auto-Refresh command */
  Command->CommandMode = FMC_SDRAM_CMD_AUTOREFRESH_MODE;
  Command->CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
  Command->AutoRefreshNumber = 8;
  Command->ModeRegisterDefinition = 0;

  /* Send the command */
  HAL_SDRAM_SendCommand(hsdram, Command, SDRAM_TIMEOUT);

  /* Step 7: Program the external memory mode register */
  tmpmrd = (uint32_t)SDRAM_MODEREG_BURST_LENGTH_1          |
                     SDRAM_MODEREG_BURST_TYPE_SEQUENTIAL   |
                     SDRAM_MODEREG_CAS_LATENCY_2           |
                     SDRAM_MODEREG_OPERATING_MODE_STANDARD |
                     SDRAM_MODEREG_WRITEBURST_MODE_SINGLE;

  Command->CommandMode = FMC_SDRAM_CMD_LOAD_MODE;
  Command->CommandTarget = FMC_SDRAM_CMD_TARGET_BANK1;
  Command->AutoRefreshNumber = 1;
  Command->ModeRegisterDefinition = tmpmrd;

  /* Send the command */
  HAL_SDRAM_SendCommand(hsdram, Command, SDRAM_TIMEOUT);

  /* Step 8: Set the refresh rate counter */
  /* (15.62 us x Freq) - 20 */
  /* Set the device refresh counter */
  //hsdram->Instance->SDRTR |= ((uint32_t)((1292)<< 1));
  hsdram->Instance->SDRTR |= ((uint32_t)((1666)<< 1));
}

/**
  * @brief  Fills buffer with user predefined data.
  * @param  pBuffer: pointer on the buffer to fill
  * @param  uwBufferLenght: size of the buffer to fill
  * @param  uwOffset: first value to fill on the buffer
  * @retval None
  */
static void Fill_Buffer(uint32_t *pBuffer, uint32_t uwBufferLenght, uint32_t uwOffset)
{
  uint32_t tmpIndex = 0;

  /* Put in global buffer different values */
  for (tmpIndex = 0; tmpIndex < uwBufferLenght; tmpIndex++ )
  {
    pBuffer[tmpIndex] = tmpIndex + uwOffset;
  }
}

/**
  * @brief  CPU L1-Cache enable.
  * @param  None
  * @retval None
  */
static void CPU_CACHE_Enable(void)
{
  /* Enable I-Cache */
  SCB_EnableICache();

  /* Enable D-Cache */
  SCB_EnableDCache();
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */

  /* Infinite loop */
  while (1)
  {
  }
}
#endif

/**
  * @}
  */

/**
  * @}
  */

